"""Computes the exact value of a belief state.  Should only be used
for small POMDPs with enumerable S, A, O spaces.  This can also be
used as an exact value iteration algorithm.
"""
import pomdp_py

def value(b, *args, horizon=1):
    """
    Computes the value of a POMDP at belief state b
    for a given pomdp. The `args` argument can be either
    a dictionary or a sequence of arguments. If it is a
    sequence of arguments, those should be in the same order
    as the arguments for `_value`, that is, S, A, Z, T, O, R, gamma.
    If it is a dictionary, then it is expected to contain S, A,
    Z, T, O, R, gamma as its fields.
    """
    if len(args) > 1:
        return _value(b, *args, horizon=horizon)
    else:
        pomdp = args[0]
        return _value(b, pomdp["S"], pomdp["A"], pomdp["Z"],
                      pomdp["T"], pomdp["O"], pomdp["R"], pomdp["gamma"],
                      horizon=horizon)

def _value(b, S, A, Z, T, O, R, gamma, horizon=1):
    """
    Computes the value of a POMDP at belief state b,
    given a POMDP defined by S, A, Z, T, O, R and gamma.

    Args:
        b (dict or Histogram): belief state, maps from every state in S to a probability
        S (list): A list of states
        A (list): A list of actions
        Z (list): A list of observations
        T (TransitionModel): The pomdp_py.TransitionModel where probability is defined
        O (ObservationModel): The pomdp_py.ObservationModel where probability is defined
        R (RewardModel): The pomdp_py.RewardModel: deterministic
        gamma (float): The discount factor
        horizon (int): The planning horizon (rewards are accumulated up
                       to the planning horizon).
    Returns:
        float: value at belief
    """
    assert type(S) == list, "S must be a list"
    assert type(A) == list, "A must be a list"
    assert type(Z) == list, "Z must be a list"
    max_qvalue = float('-inf')
    for a in A:
        # Compute Q(b,a)
        qv = qvalue(b, a, S, A, Z, T, O, R, gamma, horizon=horizon)
        if max_qvalue < qv:
            max_qvalue = qv
    return max_qvalue

def qvalue(b, a, S, A, Z, T, O, R, gamma, horizon=1):
    """Compute Q(v,a)"""
    r = expected_reward(b, R, a, T)

    expected_future_value = 0.0
    if horizon > 1:
        for o in Z:
            # compute Pr(o|b,a)*V(b')
            prob_o = belief_observation_model(o, b, a, T, O)
            # If o has non-zero probability
            if prob_o > 0:
                next_b = belief_update(b, a, o, T, O)
                next_value = value(next_b, S, A, Z, T, O, R, gamma,
                                   horizon=horizon-1)
                value_o = prob_o * next_value
                expected_future_value += value_o

    return r + gamma * expected_future_value

def expected_reward(b, R, a, T=None):
    """Returns the expected reward at a given belief"""
    r = 0.0
    for s in b:
        if T is not None:
            for sp in b:
                r += b[s] * R.sample(s, a, sp) * T.probability(sp, s, a)
        else:
            r += b[s] * R.sample(s, a, None)
    return r

def belief_observation_model(o, b, a, T, O):
    """Returns the probability of Pr(o|b,a)"""
    prob = 0.0
    for s in b:
        for sp in b:
            trans_prob = T.probability(sp, s, a)
            obsrv_prob = O.probability(o, sp, a)
            prob += obsrv_prob * trans_prob * b[s]
    return prob

def belief_update(b, a, o, T, O):
    """Returns the updated belief of `b` given
    action `a` and observation `o`."""
    next_b = {}
    total_prob = 0.0
    for sp in b:
        prob = O.probability(o, sp, a)
        trans_prob = 0.0
        for s in b:
            trans_prob += T.probability(sp, s, a) * b[s]
        prob *= trans_prob
        next_b[sp] = prob
        total_prob += next_b[sp]

    for s in next_b:
        # normalize
        next_b[s] /= total_prob
    return next_b


# Tests
def create_case(noise=0.15, init_state="tiger-left"):
    """Create a tiger problem instance with uniform belief"""
    from pomdp_problems.tiger.tiger_problem import TigerProblem, State
    tiger = TigerProblem(noise, State(init_state),
                         pomdp_py.Histogram({State("tiger-left"): 0.5,
                                             State("tiger-right"): 0.5}))
    T = tiger.agent.transition_model
    O = tiger.agent.observation_model
    S = list(T.get_all_states())
    Z = list(O.get_all_observations())
    A = list(tiger.agent.policy_model.get_all_actions())
    R = tiger.agent.reward_model
    gamma = 0.95

    b0 = tiger.agent.belief
    s0 = tiger.env.state
    return b0, s0, S, A, Z, T, O, R, gamma

def test_basic():
    b0, s0, S, A, Z, T, O, R, gamma = create_case(noise=0.15,
                                                  init_state="tiger-left")
    horizon = 3
    qvs = {}
    for a in A:
        qvs[a.name] = qvalue(b0, a, S, A, Z, T, O, R, gamma, horizon=horizon)
    assert qvs["listen"] > qvs["open-left"]
    assert qvs["listen"] > qvs["open-right"]
    assert qvs["listen"] > qvs["stay"]
    print("Pass.")


def test_planning():
    b0, s0, S, A, Z, T, O, R, gamma = create_case(noise=0.15,
                                                  init_state="tiger-left")
    horizon = 3

    # Do planning with qvalue
    b = b0
    s = s0
    print("Initial state={}. Initial belief={}".format(s, b))
    _actions = []
    for step in range(10):
        a_best = None
        maxq = float('-inf')
        for a in A:
            q = qvalue(b, a, S, A, Z, T, O, R, gamma, horizon=horizon)
            if maxq < q:
                maxq = q
                a_best = a
        sp = T.sample(s, a_best)
        o = O.sample(sp, a_best)
        b = belief_update(b, a_best, o, T, O)
        print("[Step {}] action={}, value={}, observation={},  belief={}"\
              .format(step, a_best, maxq, o, b))
        _actions.append(a_best.name)
    assert "open-right" in _actions
    print("Pass.")


if __name__ == "__main__":
    test_basic()
    test_planning()



"""Reward model for 2D Multi-object Search domain"""

import pomdp_py
from pomdp_problems.multi_object_search.domain.action import *

class MosRewardModel(pomdp_py.RewardModel):
    def __init__(self, target_objects, big=1000, small=1, robot_id=None):
        """
        robot_id (int): This model is the reward for one agent (i.e. robot),
                        If None, then this model could be for the environment.
        target_objects (set): a set of objids for target objects.
        """
        self._robot_id = robot_id
        self.big = big
        self.small = small
        self._target_objects = target_objects
        
    def probability(self, reward, state, action, next_state, normalized=False, **kwargs):
        if reward == self._reward_func(state, action):
            return 1.0
        else:
            return 0.0
        
    def sample(self, state, action, next_state,
               normalized=False, robot_id=None):
        # deterministic
        return self._reward_func(state, action, next_state, robot_id=robot_id)
    
    def argmax(self, state, action, next_state, normalized=False, robot_id=None):
        """Returns the most likely reward"""
        return self._reward_func(state, action, next_state, robot_id=robot_id)

class GoalRewardModel(MosRewardModel):
    """
    This is a reward where the agent gets reward only for detect-related actions.
    """
    def _reward_func(self, state, action, next_state, robot_id=None):
        if robot_id is None:
            assert self._robot_id is not None,\
                "Reward must be computed with respect to one robot."
            robot_id = self._robot_id
            
        reward = 0

        # If the robot has detected all objects
        if len(state.object_states[robot_id]['objects_found'])\
           == len(self._target_objects):
            return 0  # no reward or penalty; the task is finished.
        
        if isinstance(action, MotionAction):
            reward = reward - self.small - action.distance_cost
        elif isinstance(action, LookAction):
            reward = reward - self.small
        elif isinstance(action, FindAction):
            if state.object_states[robot_id]['camera_direction'] is None:
                # The robot didn't look before detect. So nothing is in the field of view.
                reward -= self.big
            else:
                # transition function should've taken care of the detection.
                new_objects_count = len(set(next_state.object_states[robot_id].objects_found)\
                                        - set(state.object_states[robot_id].objects_found))
                if new_objects_count == 0:
                    # No new detection. "detect" is a bad action.
                    reward -= self.big
                else:
                    # Has new detection. Award.
                    reward += self.big
        return reward
    



"""This file has some examples of world string."""
import random

############# Example Worlds ###########
# See env.py:interpret for definition of
# the format

world0 = (
"""
rx...
.x.xT
.....
""", "r")

world1 = (
"""
rx.T...
.x.....
...xx..
.......
.xxx.T.
.xxx...
.......
""", "r")

# Used to test the shape of the sensor
world2 = (
"""
.................
.................
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxTxxxx..
..xxxxxxrxTxxxx..
..xxxxxxxxTxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
.................
.................
""", "r")    

# Used to test sensor occlusion
world3 = (
"""
.................
.................
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxTxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxx...xxxxxx..
..xxxx..xx.xxxx..
..xxxx..r.Txxxx..
..xxxx..xx.xxxx..
..xxxxxx..xxxxx..
..xxxxTx..xxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
..xxxxxxxxxxxxx..
.................
.................
""", "r")

def random_world(width, length, num_obj, num_obstacles,
                 robot_char="r"):
    worldstr = [[ "." for i in range(width)] for j in range(length)]
    # First place obstacles
    num_obstacles_placed = 0
    while num_obstacles_placed < num_obstacles:
        x = random.randrange(0, width)
        y = random.randrange(0, length)
        if worldstr[y][x] == ".":
            worldstr[y][x] = "x"
            num_obstacles_placed += 1
            
    num_obj_placed = 0
    while num_obj_placed < num_obj:
        x = random.randrange(0, width)
        y = random.randrange(0, length)
        if worldstr[y][x] == ".":
            worldstr[y][x] = "T"
            num_obj_placed += 1

    # Finally place the robot
    while True:
        x = random.randrange(0, width)
        y = random.randrange(0, length)
        if worldstr[y][x] == ".":
            worldstr[y][x] = robot_char
            break

    # Create the string.
    finalstr = []
    for row_chars in worldstr:
        finalstr.append("".join(row_chars))
    finalstr = "\n".join(finalstr)
    return finalstr, robot_char



from pomdp_py.framework.basics cimport Action, Agent, Observation, State, Option,\
    TransitionModel, ObservationModel, RewardModel

cdef class Planner:
    """A Planner can :meth:`plan` the next action to take for a given agent (online
    planning). The planner can be updated (which may also update the agent belief)
    once an action is executed and observation received."""

    cpdef public plan(self, Agent agent):    
        """
        plan(self, Agent agent)
        The agent carries the information:
        Bt, ht, O,T,R/G, pi, necessary for planning"""
        raise NotImplementedError

    cpdef public update(self, Agent agent, Action real_action, Observation real_observation):
        """
        update(self, Agent agent, Action real_action, Observation real_observation)
        Updates the planner based on real action and observation.
        Updates the agent accordingly if necessary. If the agent's
        belief is also updated here, the `update_agent_belief`
        attribute should be set to True. By default, does nothing."""
        pass    

    def updates_agent_belief(self):
        """True if planner's update function also updates agent's
        belief."""
        return False



cdef class Distribution:
    pass
cdef class GenerativeDistribution(Distribution):
    pass
cdef class ObservationModel:
    pass
cdef class TransitionModel:
    pass
cdef class PolicyModel:
    pass
cdef class BlackboxModel:
    pass
cdef class RewardModel:
    pass

cdef class POMDP:
    cpdef public Agent agent
    cpdef public Environment env
    cpdef public str name
    
cdef class Action:
    cpdef public str name
    
cdef class State:
    pass

cdef class Observation:
    pass

cdef class Agent:
    cdef GenerativeDistribution _init_belief
    cdef PolicyModel _policy_model
    cdef TransitionModel _transition_model
    cdef RewardModel _reward_model
    cdef ObservationModel _observation_model
    cdef BlackboxModel _blackbox_model
    cdef GenerativeDistribution _cur_belief
    cdef tuple _history
    cdef dict __dict__

cdef class Environment:
    cdef State _init_state
    cdef TransitionModel _transition_model
    cdef RewardModel _reward_model
    cdef State _cur_state

cdef class Option(Action):
    pass

cpdef sample_generative_model(Agent agent, State state, Action action, float discount_factor=*)
cpdef sample_explict_models(TransitionModel T, ObservationModel O, RewardModel R, State state, Action a, float discount_factor=*)



# Defines the agent. There's nothing special
# about the MOS agent in fact, except that
# it uses models defined in ..models, and
# makes use of the belief initialization
# functions in belief.py
import pomdp_py
from .belief import *
from ..models.transition_model import *
from ..models.observation_model import *
from ..models.reward_model import *
from ..models.policy_model import *

class MosAgent(pomdp_py.Agent):
    """One agent is one robot."""
    def __init__(self,
                 robot_id,
                 init_robot_state,  # initial robot state (assuming robot state is observable perfectly)
                 object_ids,  # target object ids
                 dim,         # tuple (w,l) of the width (w) and length (l) of the gridworld search space.
                 sensor,      # Sensor equipped on the robot
                 sigma=0.01,     # parameter for observation model
                 epsilon=1,   # parameter for observation model
                 belief_rep="histogram",  # belief representation, either "histogram" or "particles".
                 prior={},       # prior belief, as defined in belief.py:initialize_belief
                 num_particles=100,  # used if the belief representation is particles
                 grid_map=None):  # GridMap used to avoid collision with obstacles (None if not provided)
        self.robot_id = robot_id
        self._object_ids = object_ids
        self.sensor = sensor

        # since the robot observes its own pose perfectly, it will have 100% prior
        # on this pose.
        prior[robot_id] = {init_robot_state.pose: 1.0}
        rth = init_robot_state.pose[2]

        # initialize belief
        init_belief = initialize_belief(dim,
                                        self.robot_id,
                                        self._object_ids,
                                        prior=prior,
                                        representation=belief_rep,
                                        robot_orientations={self.robot_id:rth},
                                        num_particles=num_particles)
        transition_model = MosTransitionModel(dim,
                                              {self.robot_id: self.sensor},
                                              self._object_ids)
        observation_model = MosObservationModel(dim,
                                                self.sensor,
                                                self._object_ids,
                                                sigma=sigma,
                                                epsilon=epsilon)
        reward_model = GoalRewardModel(self._object_ids, robot_id=self.robot_id)
        policy_model = PolicyModel(self.robot_id, grid_map=grid_map)
        super().__init__(init_belief, policy_model,
                         transition_model=transition_model,
                         observation_model=observation_model,
                         reward_model=reward_model)

    def clear_history(self):
        """Custum function; clear history"""
        self._history = None
        



import pomdp_py
from pomdp_problems.tiger.tiger_problem import TigerProblem
from pomdp_problems.tiger.tiger_problem import State as TigerState
import os
import glob

def remove_files(pattern):
    file_list = glob.glob(pattern)
    for file_path in file_list:
        os.remove(file_path)

def make_tiger(noise=0.15, init_state="tiger-left", init_belief=[0.5, 0.5]):
    """Convenient function to quickly build a tiger domain.
    Useful for testing"""
    tiger = TigerProblem(noise, TigerState(init_state),
                         pomdp_py.Histogram({TigerState("tiger-left"): init_belief[0],
                                             TigerState("tiger-right"): init_belief[1]}))
    return tiger
